#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/timeb.h>
#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include "roq.h"

SDL_Surface *screen, *video_surface, *tmp;
SDL_Surface *img;
SDL_Color *col;

unsigned char bpp_pitch;

/* -------------------------------------------------------------------------- */
void init_SDL_display(int width, int height, int full_screen)
{
	Uint32 vidflags = 0;
	
	SDL_ShowCursor(1);

	if (!SDL_WasInit(SDL_INIT_VIDEO)) 
	{	
		SDL_Init(SDL_INIT_VIDEO);
	}

#ifndef NSPIRE
	atexit(SDL_Quit);
	SDL_EventState(SDL_KEYUP, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
#endif

	screen = SDL_SetVideoMode(width, height, 16, SDL_SWSURFACE | SDL_FULLSCREEN);
	video_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 32, 0, 0, 0, 0);
	
	if (screen->format->BitsPerPixel == 32)
		bpp_pitch = 1;
	else
		bpp_pitch = 2;
	
	
	printf("%dx%dx%d  pitch: %d  Shift RGBA: %d,%d,%d,%d\n", screen->w, screen->h,
		screen->format->BitsPerPixel, screen->pitch, screen->format->Rshift,
		screen->format->Gshift, screen->format->Bshift, screen->format->Ashift);
}


/* -------------------------------------------------------------------------- */
void display_SDL(unsigned char *pa, unsigned char *pb, unsigned char *pc, int width, int height)
{

	char *buf = video_surface->pixels + screen->offset;
	int x, y, rs = (video_surface->format->Rshift/8), gs = (video_surface->format->Gshift/8),
	bs = (video_surface->format->Bshift/8);
	int r, g, b, y1, y2, u, v;

	#define limit(x) ((((x) > 0xffffff) ? 0xff0000 : (((x) <= 0xffff) ? 0 : (x) & 0xff0000)) >> 16)

	SDL_LockSurface(screen);

	for(y = 0; y < height; y++)
	{
		for(x = 0; x < width/2; x++)
		{
			y1 = *(pa++); 
			y2 = *(pa++);
			u = pb[x] - 128;
			v = pc[x] - 128;

			y1 *= 65536; y2 *= 65536;
			r = 91881 * v;
			g = -22554 * u + -46802 * v;
			b = 116130 * u;

			buf[(x*8)+rs] = limit(r + y1);
			buf[(x*8)+gs] = limit(g + y1);
			buf[(x*8)+bs] = limit(b + y1);
			buf[(x*8)+4+rs] = limit(r + y2);
			buf[(x*8)+4+gs] = limit(g + y2);
			buf[(x*8)+4+bs] = limit(b + y2);
		}
			
		if(y & 0x01) 
		{ 
			pb += width/2; 
			pc += width/2; 
		}
		buf += screen->pitch*bpp_pitch;
	}

	SDL_UnlockSurface(screen);
	
	SDL_BlitSurface(video_surface, NULL, screen, NULL);

	SDL_Flip(screen);
}

/* -------------------------------------------------------------------------- */
void usage(void)
{
	printf("SRoQPlay by Tim Ferguson.\n\n"
		"sroqplay [options] <RoQ file>\n"
		"   -f  full screen.\n"
		"\n");
}


/* -------------------------------------------------------------------------- */
void sdl_mixaudio(void *unused, Uint8 *stream, int len)
{
	roq_info *ri = (roq_info *)unused;
	static int offs, left = 0, end = 0;
	int n, pos = 0;

	if(end) return;

	if(left > 0)
	{
		pos = (left > len ? len : left);
		memcpy(stream, ri->audio + offs, pos);
		len -= pos;
		left -= pos;
		offs += pos;
	}

	while(len > 0)
	{
		if(roq_read_audio(ri) > 0)
		{
			n = (ri->audio_size > len ? len : ri->audio_size);
			memcpy(stream + pos, ri->audio, n);
			len -= n;
			pos += n;
			offs = n;
			left = ri->audio_size - n;
		}
		else 
		{ 
			end = 1; 
			break; 
		}
	}
}


/* -------------------------------------------------------------------------- */
int roqplayer(char* name)
{
	roq_info *ri;
	unsigned char done = 0;
	int ch, full_screen = 0, play_audio = 0, ret;
	
#ifdef SOUND_ENABLED
	SDL_AudioSpec fmt;
#endif
	
	Uint32 start_tm, cur_tm, display_time;

	if((ri = roq_open(name)) == NULL) return -1;
	
#ifndef NSPIRE
	printf("RoQ: length %ld:%02ld  video: %dx%d %ld frames  audio: %d channels.\n",
		ri->stream_length/60000, ((ri->stream_length + 500)/1000) % 60,
		ri->width, ri->height, ri->num_frames, ri->audio_channels);
#endif

	init_SDL_display(ri->width, ri->height, full_screen);
	
#ifdef SOUND_ENABLED
	if(ri->audio_channels > 0) play_audio = 1;

	if (play_audio)
	{
		fmt.freq = 22050;
		fmt.format = AUDIO_S16;
		fmt.channels = ri->audio_channels;
		fmt.samples = 512;
		fmt.callback = sdl_mixaudio;
		fmt.userdata = (void *)ri;

		if(SDL_OpenAudio(&fmt, NULL) < 0)
		{
			printf("Unable to open audio: %s\n", SDL_GetError());
			return -1;
		}

		SDL_PauseAudio(0);
	}
#endif

	start_tm = SDL_GetTicks();

	while(!done)
	{
		Uint8 *keystate = SDL_GetKeyState(NULL);
		
		if (keystate[SDLK_ESCAPE])
			done = 1;
	
#ifdef SOUND_ENABLED
		if (play_audio)
		{
			SDL_LockAudio();
			ret = roq_read_frame(ri);
			SDL_UnlockAudio();
		}
		else 
#endif
		ret = roq_read_frame(ri);
		
		if (ret <= 0) break;
		
		/* crappy video time sync  (should be done with audio dev) */
		
		display_time = (1000 * (Uint32)ri->frame_num)/30;
		cur_tm = SDL_GetTicks();
		
		while (cur_tm - start_tm < display_time)
		{
			SDL_Delay(1);
			cur_tm = SDL_GetTicks();
		}
		
		display_SDL(ri->y[0], ri->u[0], ri->v[0], ri->width, ri->height);
#ifndef NSPIRE
		printf("frm: %d  tm: <%d  %d>    \r", ri->frame_num, cur_tm - start_tm, display_time); fflush(stdout);
#endif
		
        SDL_Event event;
        while (SDL_PollEvent(&event))
        {
            switch (event.type)
            {
                case SDL_QUIT:
                {
                    done = 1;
                    break;
                }
            }
        }
	}
#ifdef SOUND_ENABLED
	if (play_audio) SDL_CloseAudio();
#endif
	roq_close(ri);
	SDL_FreeSurface(screen);
	SDL_FreeSurface(video_surface);
	SDL_Quit();
	return 0;
}


